/*
 * UnreachableStatesRemover.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef UNREACHABLE_STATES_REMOVER_H_
#define UNREACHABLE_STATES_REMOVER_H_

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/TA/DFTA.h>

#include <automaton/properties/ReachableStates.h>

namespace automaton {

namespace simplify {

/**
 * Algorithm for the removal of unreachable states from a finite automaton and a deterministic finite tree automaton.
 * Unreachable state is a state that is not accessible from the initial state of the automaton by any string.
 *
 * @details
 * For a finite automaton, we implement Melichar: Jazyky a překlady, 2.29.
 * For a deterministic finite tree automaton, we implement ???
 *
 * @sa automaton::simplify::Trim
 * @sa automaton::simplify::UselessStatesRemover
 */
class UnreachableStatesRemover {
public:
	/**
	 * Removes unreachable states from an automaton.
	 *
	 * @tparam T type of a finite automaton
	 *
	 * @param automaton automaton to trim
	 *
	 * @return @p automaton without unreachable states
	 */
	template < class T >
	static T remove( const T & automaton );

	/**
	 * @overload
	 * @tparam SymbolType Type for input symbols.
	 * @tparam StateType Type for states.
	 */
	template < class SymbolType, class StateType >
	static automaton::MultiInitialStateNFA < SymbolType, StateType > remove( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm );

	/**
	 * Removes unreachable states from a deterministic finite tree automaton.
	 *
	 * @overload
	 *
	 * @tparam SymbolType Type for input symbols.
	 * @tparam RankType Type for rank (arity) in ranked alphabet.
	 * @tparam StateType Type for states.
	 *
	 * @param dfta automaton to trim
	 *
	 * @return @p autoamton without unreachable states
	 */
	template < class SymbolType, class RankType, class StateType >
	static automaton::DFTA < SymbolType, RankType, StateType > remove( const automaton::DFTA < SymbolType, RankType, StateType > & dfta );

	/**
	 * Removes unreachable states from a deterministic finite tree automaton.
	 *
	 * @overload
	 *
	 * @tparam SymbolType Type for input symbols.
	 * @tparam RankType Type for rank (arity) in ranked alphabet.
	 * @tparam StateType Type for states.
	 *
	 * @param dfta automaton to trim
	 *
	 * @return @p autoamton without unreachable states
	 */
	template < class SymbolType, class RankType, class StateType >
	static automaton::NFTA < SymbolType, RankType, StateType > remove( const automaton::NFTA < SymbolType, RankType, StateType > & dfta );
};

template < class T >
T UnreachableStatesRemover::remove ( const T & fsm ) {
	using StateType = typename automaton::StateTypeOfAutomaton < T >;

	// 1a
	ext::set<StateType> Qa = automaton::properties::ReachableStates::reachableStates( fsm );

	// 2
	T M(fsm.getInitialState());

	for( const auto & q : Qa )
		M.addState( q );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fsm.getTransitions( ) )
		if( Qa.count( transition.first.first ) )
			M.addTransition( transition.first.first, transition.first.second, transition.second );

	ext::set<StateType> intersect;
	std::set_intersection( fsm.getFinalStates( ).begin(), fsm.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

template < class SymbolType, class StateType >
automaton::MultiInitialStateNFA < SymbolType, StateType > UnreachableStatesRemover::remove( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm ) {
	// 1a
	ext::set<StateType> Qa = automaton::properties::ReachableStates::reachableStates( fsm );

	// 2
	automaton::MultiInitialStateNFA < SymbolType, StateType > M;

	for( const auto & q : Qa )
		M.addState( q );

	M.setInitialStates( fsm.getInitialStates() );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fsm.getTransitions( ) )
		if( Qa.count( transition.first.first ) )
			M.addTransition( transition.first.first, transition.first.second, transition.second );

	ext::set<StateType> intersect;
	std::set_intersection( fsm.getFinalStates( ).begin(), fsm.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

template < class SymbolType, class RankType, class StateType >
automaton::DFTA < SymbolType, RankType, StateType > UnreachableStatesRemover::remove( const automaton::DFTA < SymbolType, RankType, StateType > & fta ) {
	// 1a
	ext::set<StateType> Qa = automaton::properties::ReachableStates::reachableStates( fta );

	// 2
	automaton::DFTA < SymbolType, RankType, StateType > M;

	for( const auto & q : Qa )
		M.addState( q );

	for( const auto & a : fta.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fta.getTransitions( ) )
		if( std::all_of ( transition.first.second.begin ( ), transition.first.second.end ( ), [ & ] ( const StateType & state ) { return Qa.count ( state ); } ) )
			M.addTransition ( transition.first.first, transition.first.second, transition.second );

	ext::set<StateType> intersect;
	std::set_intersection( fta.getFinalStates( ).begin(), fta.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

template < class SymbolType, class RankType, class StateType >
automaton::NFTA < SymbolType, RankType, StateType > UnreachableStatesRemover::remove( const automaton::NFTA < SymbolType, RankType, StateType > & fta ) {
	// 1a
	ext::set<StateType> Qa = automaton::properties::ReachableStates::reachableStates( fta );

	// 2
	automaton::NFTA < SymbolType, RankType, StateType > M;

	for( const auto & q : Qa )
		M.addState( q );

	for( const auto & a : fta.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fta.getTransitions( ) )
		if( std::all_of ( transition.first.second.begin ( ), transition.first.second.end ( ), [ & ] ( const StateType & state ) { return Qa.count ( state ); } ) )
			M.addTransitions ( transition.first.first, transition.first.second, transition.second );

	ext::set<StateType> intersect;
	std::set_intersection( fta.getFinalStates( ).begin(), fta.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* UNREACHABLE_STATES_REMOVER_H_ */
